/*
	GameCube Save Util by CrowTRobo
	Transfers GC memory card saves between GC and PC.
	Started October 16, 2003.
*/

#ifdef LINUX
#include <errno.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <tcpd.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#else
#include <winsock.h>
#include <conio.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "crc32.h"

#define DATA_LEN 1024
#define CMD_LEN 17

#define CMD_OFFS 4
#define DATA_OFFS CMD_LEN+CMD_OFFS

#ifdef LINUX
#define SOCKET_ERROR -1
#define INVALID_SOCKET -1
#define WSAEWOULDBLOCK EWOULDBLOCK
#define stricmp strcasecmp
#define strnicmp strncasecmp
#define WSACleanup() close_keyboard();
#define closesocket(s) close(s)
#define SOCKET int
#define BOOL int
#define TRUE 1
#define FALSE 0
#endif

#define SOCKET_OK 1
#define SOCKET_CANCEL 2
#define SOCKET_CRCERR 3

#define NONE 0
#define GETLIST 1
#define BACKUPFILE 2
#define RESTOREFILE 3
#define BACKUPRAW 4
#define RESTORERAW 5
#define CONFIG 6

typedef struct {
	unsigned long crc;
	char cmd[CMD_LEN];
	unsigned char data[DATA_LEN];
} packet;

SOCKET sock;
struct sockaddr_in client;
FILE *fp = NULL;
BOOL connected = FALSE;
unsigned char *buffer = NULL;
int client_len;

#ifdef LINUX
static struct termios initial_settings, new_settings;
static int peek_character = -1;

void init_keyboard() {

	tcgetattr(0,&initial_settings);
	new_settings = initial_settings;
	new_settings.c_lflag &= ~ICANON;
	new_settings.c_lflag &= ~ECHO;
	new_settings.c_lflag &= ~ISIG;
	new_settings.c_cc[VMIN] = 1;
	new_settings.c_cc[VTIME] = 0;
	tcsetattr(0, TCSANOW, &new_settings);
}

void close_keyboard() {

	tcsetattr(0, TCSANOW, &initial_settings);
}

int kbhit() {
 
	unsigned char ch;
	int nread;
 
	if (peek_character != -1) 
		return 1;
		
	new_settings.c_cc[VMIN]=0;
	tcsetattr(0, TCSANOW, &new_settings);
	nread = read(0,&ch,1);
	new_settings.c_cc[VMIN]=1;
	tcsetattr(0, TCSANOW, &new_settings);
		
	if(nread == 1) {
		peek_character = ch;
		return 1;
	}
	
	return 0;
}

int getch() {

	char ch;
	
	if(peek_character != -1) {
		ch = peek_character;
		peek_character = -1;
		return ch;
	}
	
	read(0,&ch,1);
	return ch;
}
#endif

unsigned long CRC32(unsigned char *data, int size) {

	unsigned long crc = 0xffffffff;
	int i;

	for (i = 0; i < size; i++)
    crc = (crc >> 8) ^ crc32_lut[(crc & 0xff) ^ data[i]];

  return crc ^ 0xffffffff;
}

void PutPacket(packet *pkt, unsigned char *buf) {

  unsigned long val;
  
	memcpy(&buf[CMD_OFFS], pkt->cmd, CMD_LEN);
	memcpy(&buf[DATA_OFFS], pkt->data, DATA_LEN);

	pkt->crc = CRC32(&buf[CMD_OFFS], CMD_LEN+DATA_LEN);
	val = htonl(pkt->crc);
	memcpy(buf, &val, 4);
}

unsigned long GetPacket(packet *pkt, unsigned char *buf) {

	unsigned long val;

	memcpy(&val, buf, 4);
	pkt->crc = ntohl(val);
	memcpy(&pkt->cmd, &buf[CMD_OFFS], CMD_LEN);
	memcpy(&pkt->data, &buf[DATA_OFFS], DATA_LEN);

	return CRC32(&buf[CMD_OFFS], CMD_LEN+DATA_LEN);
}

void PutLong(unsigned long val, unsigned char *buffer) {

	buffer[0] = (unsigned char)(val & 0xff);
	buffer[1] = (unsigned char)((val >> 8) & 0xff);
	buffer[2] = (unsigned char)((val >> 16) & 0xff);
	buffer[3] = (unsigned char)((val >> 24) & 0xff);
}

unsigned long GetLong(unsigned char *buffer) {

	unsigned long val;

	val = buffer[0];
	val |= (buffer[1] << 8);
	val |= (buffer[2] << 16);
	val |= (buffer[3] << 24);

	return val;
}

int RecvPacket(packet *pkt) {

  unsigned long crc_in;
  int ret;

  while (1) {

		if (kbhit()) {
			getch();
			return SOCKET_CANCEL;
		}
    
    ret = recvfrom(sock, (char *)buffer, sizeof(packet), 0, (struct sockaddr *)&client, &client_len);
			
    if (ret == SOCKET_ERROR) {
#ifdef LINUX
			ret = errno;
#else
			ret = WSAGetLastError();
#endif

			if (ret == WSAEWOULDBLOCK)
				continue;
			else {
				printf("Error receiving packet.\n");
				return SOCKET_ERROR;
			}
		}

    if (ret != sizeof(packet)) {
      printf("Only received %d out of %d bytes", ret, sizeof(packet));
      return SOCKET_ERROR;
    }

    crc_in = GetPacket(pkt, buffer);
			
    if (crc_in != pkt->crc) {
      printf("CRC error in packet.\n");
      return SOCKET_CRCERR;
    }

    break;
  }

  return SOCKET_OK;
}

int SendPacket(packet *pkt) {

  int ret;

	while (1) {

		if (kbhit()) {
			getch();
			return SOCKET_CANCEL;
		}

		PutPacket(pkt, buffer);

		ret = sendto(sock, (char *)buffer, sizeof(packet), 0, (struct sockaddr *)&client, client_len);

		if (ret == SOCKET_ERROR) {
#ifdef LINUX
			ret = errno;
#else
			ret = WSAGetLastError();
#endif

			if (ret == WSAEWOULDBLOCK)
				continue;
			else {
				printf("Error sending packet.\n");
				return SOCKET_ERROR;
			}
		}

		if (ret != sizeof(packet)) {
      printf("Only received %d out of %d bytes", ret, sizeof(packet));
      return SOCKET_ERROR;
    }

		break;
	}

  return SOCKET_OK;
}

int Connect(packet *pkt) {

  int ret;

	memset(pkt, 0, sizeof(packet));
	strcpy(pkt->cmd, "CONNECT");

  ret = SendPacket(pkt);

	if (ret != SOCKET_OK)
    return ret;

  ret = RecvPacket(pkt);

	if (ret != SOCKET_OK)
		return ret;

	if (strcmp(pkt->cmd, "ACK") != 0)
		return SOCKET_ERROR;

	memset(pkt, 0, sizeof(packet));
	strcpy(pkt->cmd, "ACK");

  return SendPacket(pkt);
}

int GetList(packet *pkt, int slot) {

	packet pkt2;
	int ret, end = 0;

	memset(pkt, 0, sizeof(packet));
	strcpy(pkt->cmd, "GETLIST");
	PutLong(slot, pkt->data);
	ret = SendPacket(pkt);

	if (ret != SOCKET_OK)
		return ret;

	printf("\n");

	while (1) {
		ret = RecvPacket(pkt);

		if (ret != SOCKET_OK)
			return ret;

		memcpy(&pkt2, pkt, sizeof(packet));

		if (strcmp(pkt->cmd, "ENDGETLIST") == 0)
			end = 1;

		memset(pkt, 0, sizeof(packet));
		strcpy(pkt->cmd, "OK");
		ret = SendPacket(pkt);

		if (ret != SOCKET_OK)
			return ret;

		ret = RecvPacket(pkt);

		if (ret != SOCKET_OK)
			return ret;

		if (strcmp(pkt->cmd, "OK") != 0)
			continue;

		if (end == 1)
			break;

		printf("%3ld ", GetLong(&pkt2.data[0]));
		printf("%-4s ", &pkt2.data[4]);
		printf("%-2s ", &pkt2.data[9]);
		printf("%-32s ", &pkt2.data[12]);
		printf("%ld\n", GetLong(&pkt2.data[50]));
	}

	return SOCKET_OK;
}

int RecvFilePart(packet *pkt, unsigned char *data, int offset, int len) {

	int ret;

	while (1) {
		ret = RecvPacket(pkt);

		if (ret == SOCKET_CRCERR) {
			memset(pkt, 0, sizeof(packet));
			strcpy(pkt->cmd, "ERROR");
			ret = SendPacket(pkt);

			if (ret != SOCKET_OK)
				return ret;

			continue;
		}
		else if (ret != SOCKET_OK)
			return ret;

		memcpy(&data[offset], pkt->data, len);

		memset(pkt, 0, sizeof(packet));
		strcpy(pkt->cmd, "OK");
		ret = SendPacket(pkt);

		if (ret != SOCKET_OK)
			return ret;

		break;
	}

	return SOCKET_OK;
}

int SendFilePart(packet *pkt, unsigned char *data, int offset, int len) {

	int ret;

	while (1) {
		memset(pkt, 0, sizeof(packet));
		strcpy(pkt->cmd, "PART");
		memcpy(pkt->data, &data[offset], len);
		ret = SendPacket(pkt);

		if (ret != SOCKET_OK)
			return ret;

		ret = RecvPacket(pkt);

		if (ret != SOCKET_OK)
			return ret;

		if (strcmp(pkt->cmd, "OK") == 0)
			break;
	}

	return SOCKET_OK;
}

int BackupFile(packet *pkt, int slot, int filenum) {

	unsigned char *data;
	int i, ret, size;

	memset(pkt, 0, sizeof(packet));

	if (filenum >= 0) {
		strcpy(pkt->cmd, "BACKUPFILE");
		PutLong(slot, pkt->data);
		PutLong(filenum, &pkt->data[4]);
	}
	else {
		strcpy(pkt->cmd, "BACKUPRAW");
		PutLong(slot, pkt->data);
	}

	ret = SendPacket(pkt);

	if (ret != SOCKET_OK)
		return ret;

	ret = RecvPacket(pkt);

	if (ret != SOCKET_OK)
		return ret;

	if (strcmp(pkt->cmd, "OK") != 0) {
		printf("An error occured on the GameCube.\n");
		return SOCKET_ERROR;
	}

	size = GetLong(pkt->data);

	if ((data = (unsigned char *)malloc(size)) == NULL) {
		printf("Could not allocate %d bytes of data.\n", size);
		
		memset(pkt, 0, sizeof(packet));
		strcpy(pkt->cmd, "ERROR");
		return SendPacket(pkt);
	}

	memset(data, 0, size);
	memset(pkt, 0, sizeof(packet));
	strcpy(pkt->cmd, "OK");
	ret = SendPacket(pkt);

	if (ret != SOCKET_OK) {
		free(data);
		return ret;
	}

	printf("Receiving file\n");

	for (i = 0; i < (size / DATA_LEN); i++) {
		ret = RecvFilePart(pkt, data, i*DATA_LEN, DATA_LEN);

		if (ret != SOCKET_OK) {
			free(data);
			return ret;
		}
	}

	if (size % DATA_LEN) {
		ret = RecvFilePart(pkt, data, i*DATA_LEN, size % DATA_LEN);

		if (ret != SOCKET_OK) {
			free(data);
			return ret;
		}
	}

	fwrite(data, 1, size, fp);
	free(data);
	printf("Successfully received file.\n");

	return SOCKET_OK;
}

int RestoreFile(packet *pkt, int slot, int command) {

	unsigned char *data;
	int i, ret, size;

	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	if ((data = (unsigned char *)malloc(size)) == NULL) {
		printf("Could not allocate %d bytes of data.\n", size);
		return SOCKET_ERROR;
	}

	fread(data, 1, size, fp);

	memset(pkt, 0, sizeof(packet));

	if (command == RESTOREFILE)
		strcpy(pkt->cmd, "RESTOREFILE");
	else
		strcpy(pkt->cmd, "RESTORERAW");

	PutLong(slot, pkt->data);
	PutLong(size, &pkt->data[4]);
	ret = SendPacket(pkt);

	if (ret != SOCKET_OK) {
		free(data);
		return ret;
	}

	ret = RecvPacket(pkt);

	if (ret != SOCKET_OK) {
		free(data);
		return ret;
	}

	if (strcmp(pkt->cmd, "OK") != 0) {
		printf("An error occured on the GameCube.\n");
		free(data);
		return SOCKET_ERROR;
	}

	printf("Sending file\n");

	for (i = 0; i < (size / DATA_LEN); i++) {
		ret = SendFilePart(pkt, data, i*DATA_LEN, DATA_LEN);

		if (ret != SOCKET_OK) {
			free(data);
			return ret;
		}
	}

	if (size % DATA_LEN) {
		ret = SendFilePart(pkt, data, i*DATA_LEN, size % DATA_LEN);

		if (ret != SOCKET_OK) {
			free(data);
			return ret;
		}
	}

	free(data);
	printf("Successfully sent file.\n");

	return SOCKET_OK;
}

unsigned long FindIpstr() {

	char c[2], str[8];
	int i = 0;

	while (!feof(fp)) {
		c[0] = fgetc(fp);
		c[1] = '\0';

		if (i) {

			if (strpbrk(c, "PSTART") != NULL)
				str[i++] = c[0];
			else 
				i = 0;

			if (i == 7) {
				str[i] = '\0';

				if (strcmp(str, "IPSTART") == 0)
					return ftell(fp);
				else
					i = 0;
			}
		}

		if (c[0] == 'I') 
			str[i++] = c[0];
	}

	return 0;
}

int ParseCfg(unsigned long *pcip, unsigned long *gcip) {

	char buf[128], *tok, *tok2;

	*pcip = 0;
	*gcip = 0;

	while (!feof(fp)) {
		fgets(buf, 128, fp);

		if ((tok = strtok(buf, " =")) == NULL)
			continue;

		if ((tok2 = strtok(NULL, " =")) == NULL)
			continue;

		if (stricmp(tok, "pcip") == 0)
			*pcip = inet_addr(tok2);
		else if (stricmp(tok, "gcip") == 0) 
			*gcip = inet_addr(tok2);
	}

	if (*pcip == 0 || *gcip == 0)
		return 0;

	return 1;
}

#ifdef LINUX
void Exit(int sig) {

	if (buffer != NULL)
		free(buffer);
	
	if (fp != NULL)
		fclose(fp);
	
	closesocket(sock);
	close_keyboard();
	signal(sig, SIG_DFL);
	raise(sig);
}
#else
BOOL WINAPI Exit(DWORD dwCtrlType) {

	if (buffer != NULL)
		free(buffer);

	if (fp != NULL)
		fclose(fp);

	closesocket(sock);
	WSACleanup();

	return FALSE;
}
#endif

int main(int argc, char* argv[]) {

#ifndef LINUX
	WSADATA wsadata;
#endif
	struct sockaddr_in addr;
	packet pkt;
	unsigned char ipdata[8];
	unsigned long ioctl_arg, pcip, gcip, offset;
	int i, ret, command = NONE, slot = -1, filenum;

	printf("\nGCS 1.0 by CrowTRobo\n\n");

	if (argc < 2) {
		printf("ctr-gcs command slot <file>\n\n");
		printf("commands:\n");
		printf("\t-l\t List files on memcard\n");
		printf("\t-b#\t Backup file # to <file>\n");
		printf("\t-r\t Restore <file> to memcard\n");
		printf("\t-ib\t Backup raw image of memcard\n");
		printf("\t-ir\t Restore raw image of memcard\n");
		printf("\t-c\t Configure DOL with IP addresses\n"); 
		printf("\nslot:\n");
		printf("\tA\t Work with memcard in slot A\n");
		printf("\tB\t Work with memcard in slot B\n");
		return 1;
	}

	for (i = 1; i < 3; i++) {

		if (argv[i] == NULL)
			break;

		if (argv[i][0] == '-') {

			if (stricmp(&argv[i][1], "l") == 0)
				command = GETLIST;
			else if (strnicmp(&argv[i][1], "b", 1) == 0) {
				command = BACKUPFILE;
				filenum = atoi(&argv[i][2]);

				if (filenum < 0 || filenum >= 127) {
					printf("Invalid file number: %d\n", filenum);
					return 1;
				}
			}
			else if (strcmp(&argv[i][1], "r") == 0)
				command = RESTOREFILE;
			else if (stricmp(&argv[i][1], "ib") == 0)
				command = BACKUPRAW;
			else if (stricmp(&argv[i][1], "ir") == 0)
				command = RESTORERAW;
			else if (stricmp(&argv[i][1], "c") == 0)
				command = CONFIG;			
			else {
				printf("Invalid command: %s\n", argv[i]);
				return 1;
			}
		}
		else {

			if (stricmp(argv[i], "A") == 0)
				slot = 0;
			else if (stricmp(argv[i], "B") == 0)
				slot = 1;
		}
	}

	if (command == NONE) {
		printf("No command was given.\n");
		return 1;
	}
	
#ifdef LINUX
	init_keyboard();
#else
	ret = WSAStartup(MAKEWORD(2,0), &wsadata);

	if (ret) {
		printf("Could not initialize winsock.\n");
		return 1;
	}
#endif

	if ((fp = fopen("ctr-gcs.cfg", "rt")) == NULL) {
		printf("Could not open gcs.cfg for reading.\n");
		WSACleanup();
		return 1;
	}

	ret = ParseCfg(&pcip, &gcip);
	fclose(fp);
	fp = NULL;

	if (!ret) {
		printf("Failed to read IP addresses from the config file.\n");
		WSACleanup();
		return 1;
	}

	addr.sin_addr.s_addr = pcip;
	printf("PC IP = %s\n", inet_ntoa(addr.sin_addr));
	addr.sin_addr.s_addr = gcip;
	printf("GC IP = %s\n\n", inet_ntoa(addr.sin_addr));

	if (command == CONFIG) {

		if ((fp = fopen("ctr-gcs.dol", "rb+")) == NULL) {
			printf("Could not open gcs.dol for config.\n");
			WSACleanup();
			return 1;
		}

		offset = FindIpstr();
		
		if (offset == 0) {
			printf("gcs.dol is not the proper DOL file.\n");
			fclose(fp);
			WSACleanup();
			return 1;
		}

		PutLong(ntohl(pcip), ipdata);
		PutLong(ntohl(gcip), &ipdata[4]);
		fseek(fp, offset, SEEK_SET);
		fwrite(ipdata, 1, 8, fp);
		fclose(fp);
		printf("gcs.dol configured.\n");
		WSACleanup();
		return 0;
	}	

	if (slot == -1) {
		printf("A valid slot was not given.\n");
		WSACleanup();
		return 1;
	}
	
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET) {
		printf("Could not create socket.\n");
		WSACleanup();
		return 1;
	}

#ifdef LINUX
	signal(SIGINT, Exit);
#else
	SetConsoleCtrlHandler(Exit, TRUE);
#endif

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(9002);
	addr.sin_addr.s_addr = pcip;

	memset(&client, 0, sizeof(client));
	client.sin_family = AF_INET;
	client.sin_port = htons(9002);
	client.sin_addr.s_addr = gcip;
	client_len = sizeof(client);

	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == SOCKET_ERROR) {
		printf("Could not bind socket.\n");
		closesocket(sock);
		WSACleanup();
		return 1;
	}

#ifdef LINUX
	ioctl_arg = fcntl(sock, F_GETFL);
	ret = fcntl(sock, F_SETFL, ioctl_arg | O_NONBLOCK);	
#else
	ioctl_arg = 1;
	ret = ioctlsocket(sock, FIONBIO, &ioctl_arg);
#endif

	if (ret == SOCKET_ERROR) {
		printf("Could not set socket to non-blocking.\n");
		closesocket(sock);
		WSACleanup();
		return 1;
	}

	if ((buffer = (unsigned char *)malloc(sizeof(packet))) == NULL) {
		printf("Could not allocate memory for buffer.\n");
		closesocket(sock);
		WSACleanup();
		return 1;
	}
	
	if (command >= BACKUPFILE && command <= RESTORERAW) {

		if (argv[i] == NULL) {
			printf("No filename was given.\n");
			free(buffer);
			closesocket(sock);
			WSACleanup();
			return 1;
		}

		if (command == BACKUPFILE || command == BACKUPRAW) {

			if ((fp = fopen(argv[i], "wb")) == NULL) {
				printf("Could not open %s for writing.\n", argv[i]);
				free(buffer);
				closesocket(sock);
				WSACleanup();
				return 1;
			}
		}
		else if (command == RESTOREFILE || command == RESTORERAW) {

			if ((fp = fopen(argv[i], "rb")) == NULL) {
				printf("Could not open %s for reading.\n", argv[i]);
				free(buffer);
				closesocket(sock);
				WSACleanup();
				return 1;
			}
		}
	}

	printf("Waiting to connect\n");

	while (1) {
  
    if (kbhit()) {
			getch();
			break;
		}

    if (connected == FALSE) {
			ret = Connect(&pkt);

			if (ret == SOCKET_OK) {
				printf("Connected\n");
				connected = TRUE;
			}
			else if (ret == SOCKET_CANCEL)
				break;
    }
		else {

			if (command == GETLIST) {
				GetList(&pkt, slot);
				break;
			}
			else if (command == BACKUPFILE) {
				BackupFile(&pkt, slot, filenum);
				break;
			}
			else if (command == RESTOREFILE) {
				RestoreFile(&pkt, slot, RESTOREFILE);
				break;
			}
			else if (command == BACKUPRAW) {
				BackupFile(&pkt, slot, -1);
				break;
			}
			else if (command == RESTORERAW) {
				RestoreFile(&pkt, slot, RESTORERAW);
				break;
			}
		}
  }

	if (fp != NULL)
		fclose(fp);

	free(buffer);
	closesocket(sock);
	WSACleanup();

	return 0;
}
